//includes
#include <SdFat.h>// for SDFat library see: https://github.com/greiman/SdFat


//functions
boolean getUserSerial();
boolean getDate();
boolean getIntro();

void burnEEPROM(byte EEPROMdata);
void getString();
void readALlFromChip();
void showHex (const byte b, const boolean newline = false, const boolean show0x = true);
void blink (const int whichLED1,
            const int whichLED2,
            const byte times = 1,
            const unsigned long repeat = 1,
            const unsigned long interval = 200);
void ShowMessage (const byte which);
byte BB_SPITransfer (byte c);
byte program (const byte b1, const byte b2 = 0, const byte b3 = 0, const byte b4 = 0);
byte readFlash (unsigned long addr);
void writeFlash (unsigned long addr, const byte data);
bool hexConv (const char * (& pStr), byte & b);
void pollUntilReady ();
void showProgress ();
void clearPage ();
void commitPage (unsigned long addr);
void writeData (const unsigned long addr, const byte * pData, const int length);
void verifyData (const unsigned long addr, const byte * pData, const int length);
bool processLine (const char * pLine, const byte action);
bool readHexFile (const char * fName, const byte action);
bool startProgramming ();
void stopProgramming ();
void getSignature ();
void getFuseBytes ();
void writeFuse (const byte newValue, const byte instruction);
bool updateFuses (const bool writeIt);
bool chooseInputFile ();
bool writeFlashContents ();



//variables and constants
const unsigned int ENTER_PROGRAMMING_ATTEMPTS = 2;
String inputstring = "";
boolean input_string_complete = false;
byte EEPROMindex = 0;
const bool allowTargetToRun = true;  // if true, programming lines are freed when not programming
const char dateMessage[] PROGMEM  = {"Please Enter Date like this: MM/DD/YY \n Example for July 4th, 2018, would be 07/04/18\n"};
const char introMessage[]  PROGMEM  = {"What do you want to do?\n  R = Read Board Information\n  W = Write parameters\n  N = Brand New Board Programming\n"};
const char fuseBurn[]  PROGMEM  = {"Burning Fuse Config:  Should be LFuse=0xFF, HFuse=0xDF, EFuse=0xFD\n"};
const char wrongLength[]  PROGMEM  = {"Sorry wrong length\n"};
// fixed file name to read from SD card (root directory)
const char wantedFile [] = "firmware.hex";

// which switch to close to start programming the target chip
const byte startSwitch = 2;

// the three "status" LEDs
const int errorLED = A0;
const int readyLED = A1;
const int workingLED = A4;
const int buzzerPin = 3;

const int noLED = -1;  // must be int! - can be negative

// status "messages"
typedef enum {
  MSG_NO_SD_CARD,          // cannot open SD card
  MSG_CANNOT_OPEN_FILE,    // canoot open file 'wantedFile' (above)
  MSG_LINE_TOO_LONG,       // line on disk too long to read
  MSG_LINE_TOO_SHORT,      // line too short to be valid
  MSG_LINE_DOES_NOT_START_WITH_COLON,  // line does not start with a colon
  MSG_INVALID_HEX_DIGITS,  // invalid hex where there should be hex
  MSG_BAD_SUMCHECK,        // line fails sumcheck
  MSG_LINE_NOT_EXPECTED_LENGTH,  // record not length expected
  MSG_UNKNOWN_RECORD_TYPE,  // record type not known
  MSG_NO_END_OF_FILE_RECORD,  // no 'end of file' at end of file
  MSG_FILE_TOO_LARGE_FOR_FLASH,  // file will not fit into flash

  MSG_CANNOT_ENTER_PROGRAMMING_MODE,  // cannot program target chip
  MSG_NO_BOOTLOADER_FUSE,             // chip does not have bootloader
  MSG_CANNOT_FIND_SIGNATURE,      // cannot find chip signature
  MSG_UNRECOGNIZED_SIGNATURE,     // signature not known
  MSG_BAD_START_ADDRESS,          // file start address invalid
  MSG_VERIFICATION_ERROR,         // verification error after programming
  MSG_FLASHED_OK,                 // flashed OK
} msgType;

//SPI to target pins
const byte MSPIM_SCK = 4;  // port D bit 4
const byte MSPIM_SS  = 5;  // port D bit 5
const byte BB_MISO   = 6;  // port D bit 6
const byte BB_MOSI   = 7;  // port D bit 7
// 8 MHz clock on this pin
const byte CLOCKOUT = 9;

#define BB_MISO_PORT PIND
#define BB_MOSI_PORT PORTD
#define BB_SCK_PORT PORTD
const byte BB_SCK_BIT = 4;
const byte BB_MISO_BIT = 6;
const byte BB_MOSI_BIT = 7;

// control speed of programming
const byte BB_DELAY_MICROSECONDS = 6;
// target board reset goes to here
const byte RESET = MSPIM_SS;
// SD chip select pin
const uint8_t chipSelect = SS;
const unsigned long NO_PAGE = 0xFFFFFFFF;
const int MAX_FILENAME = 13;

// actions to take
enum {
  checkFile,
  verifyFlash,
  writeToFlash,
};

// file system object
SdFat sd;

// copy of fuses/lock bytes found for this processor
byte fuses [5];

// meaning of bytes in above array
enum {
  lowFuse,
  highFuse,
  extFuse,
  lockByte,
  calibrationByte
};

// structure to hold signature and other relevant data about each chip
typedef struct {
  byte sig [3];
  const char * desc;
  unsigned long flashSize;
  unsigned int baseBootSize;
  unsigned long pageSize;     // bytes
  byte fuseWithBootloaderSize;  // ie. one of: lowFuse, highFuse, extFuse
  byte timedWrites;    // if pollUntilReady won't work by polling the chip
} signatureType;

const unsigned long kb = 1024;
const byte NO_FUSE = 0xFF;

// see Atmega datasheets
const signatureType signatures [] PROGMEM =
{
  //     signature        description   flash size   bootloader  flash  fuse
  //                                                     size    page    to
  //                                                             size   change

  // Attiny84 family
  { { 0x1E, 0x91, 0x0B }, "ATtiny24",   2 * kb,           0,   32,   NO_FUSE },
  { { 0x1E, 0x92, 0x07 }, "ATtiny44",   4 * kb,           0,   64,   NO_FUSE },
  { { 0x1E, 0x93, 0x0C }, "ATtiny84",   8 * kb,           0,   64,   NO_FUSE },

  // Attiny85 family
  { { 0x1E, 0x91, 0x08 }, "ATtiny25",   2 * kb,           0,   32,   NO_FUSE },
  { { 0x1E, 0x92, 0x06 }, "ATtiny45",   4 * kb,           0,   64,   NO_FUSE },
  { { 0x1E, 0x93, 0x0B }, "ATtiny85",   8 * kb,           0,   64,   NO_FUSE },

  // Atmega328 family
  { { 0x1E, 0x92, 0x0A }, "ATmega48PA",   4 * kb,         0,    64,  NO_FUSE },
  { { 0x1E, 0x93, 0x0F }, "ATmega88PA",   8 * kb,       256,   128,  extFuse },
  { { 0x1E, 0x94, 0x0B }, "ATmega168PA", 16 * kb,       256,   128,  extFuse },
  { { 0x1E, 0x95, 0x0F }, "ATmega328P",  32 * kb,       512,   128,  highFuse },

  // Atmega644 family
  { { 0x1E, 0x94, 0x0A }, "ATmega164P",   16 * kb,      256,   128,  highFuse },
  { { 0x1E, 0x95, 0x08 }, "ATmega324P",   32 * kb,      512,   128,  highFuse },
  { { 0x1E, 0x96, 0x0A }, "ATmega644P",   64 * kb,   1 * kb,   256,  highFuse },

  // Atmega2560 family
  { { 0x1E, 0x96, 0x08 }, "ATmega640",    64 * kb,   1 * kb,   256,  highFuse },
  { { 0x1E, 0x97, 0x03 }, "ATmega1280",  128 * kb,   1 * kb,   256,  highFuse },
  { { 0x1E, 0x97, 0x04 }, "ATmega1281",  128 * kb,   1 * kb,   256,  highFuse },
  { { 0x1E, 0x98, 0x01 }, "ATmega2560",  256 * kb,   1 * kb,   256,  highFuse },

  { { 0x1E, 0x98, 0x02 }, "ATmega2561",  256 * kb,   1 * kb,   256,  highFuse },

  // AT90USB family
  { { 0x1E, 0x93, 0x82 }, "At90USB82",    8 * kb,       512,   128,  highFuse },
  { { 0x1E, 0x94, 0x82 }, "At90USB162",  16 * kb,       512,   128,  highFuse },

  // Atmega32U2 family
  { { 0x1E, 0x93, 0x89 }, "ATmega8U2",    8 * kb,       512,   128,  highFuse  },
  { { 0x1E, 0x94, 0x89 }, "ATmega16U2",  16 * kb,       512,   128,  highFuse  },
  { { 0x1E, 0x95, 0x8A }, "ATmega32U2",  32 * kb,       512,   128,  highFuse  },

  // Atmega32U4 family -  (datasheet is wrong about flash page size being 128 words)
  { { 0x1E, 0x94, 0x88 }, "ATmega16U4",  16 * kb,       512,   128,  highFuse },
  { { 0x1E, 0x95, 0x87 }, "ATmega32U4",  32 * kb,       512,   128,  highFuse },

  // ATmega1284P family
  { { 0x1E, 0x97, 0x05 }, "ATmega1284P", 128 * kb,   1 * kb,   256,  highFuse  },

  // ATtiny4313 family
  { { 0x1E, 0x91, 0x0A }, "ATtiny2313A",   2 * kb,        0,    32,  NO_FUSE  },
  { { 0x1E, 0x92, 0x0D }, "ATtiny4313",    4 * kb,        0,    64,  NO_FUSE  },

  // ATtiny13 family
  { { 0x1E, 0x90, 0x07 }, "ATtiny13A",     1 * kb,        0,    32,  NO_FUSE },

  // Atmega8A family
  { { 0x1E, 0x93, 0x07 }, "ATmega8A",      8 * kb,      256,    64,  highFuse, true },

  // ATmega64rfr2 family
  { { 0x1E, 0xA6, 0x02 }, "ATmega64rfr2",  256 * kb, 1 * kb,   256, highFuse },
  { { 0x1E, 0xA7, 0x02 }, "ATmega128rfr2", 256 * kb, 1 * kb,   256, highFuse },
  { { 0x1E, 0xA8, 0x02 }, "ATmega256rfr2", 256 * kb, 1 * kb,   256, highFuse },

};  // end of signatures

char name[MAX_FILENAME] = { 0 };  // current file name

// number of items in an array
#define NUMITEMS(arg) ((unsigned int) (sizeof (arg) / sizeof (arg [0])))

// programming commands to send via SPI to the chip
enum {
  progamEnable = 0xAC,

  // writes are preceded by progamEnable
  chipErase = 0x80,
  writeLockByte = 0xE0,
  writeLowFuseByte = 0xA0,
  writeHighFuseByte = 0xA8,
  writeExtendedFuseByte = 0xA4,

  pollReady = 0xF0,

  programAcknowledge = 0x53,

  readSignatureByte = 0x30,
  readCalibrationByte = 0x38,

  readLowFuseByte = 0x50,       readLowFuseByteArg2 = 0x00,
  readExtendedFuseByte = 0x50,  readExtendedFuseByteArg2 = 0x08,
  readHighFuseByte = 0x58,      readHighFuseByteArg2 = 0x08,
  readLockByte = 0x58,          readLockByteArg2 = 0x00,

  readProgramMemory = 0x20,
  writeProgramMemory = 0x4C,
  loadExtendedAddressByte = 0x4D,
  loadProgramMemory = 0x40,
  readEEPROMbyte = 0xA0,
  writeEEPROMbyte = 0xC0

                    /*
                       0xC0 WRITE BYTE TO EEPROM
                      0xA0 READ BYTE FROM EEPROM

                    */

};  // end of enum

// which program instruction writes which fuse
const byte fuseCommands [4] = { writeLowFuseByte, writeHighFuseByte, writeExtendedFuseByte, writeLockByte };

// types of record in .hex file
enum {
  hexDataRecord,  // 00
  hexEndOfFile,   // 01
  hexExtendedSegmentAddressRecord, // 02
  hexStartSegmentAddressRecord,  // 03
  hexExtendedLinearAddressRecord, // 04
  hexStartLinearAddressRecord // 05
};
// if signature found in above table, this is its index
int foundSig = -1;
byte lastAddressMSB = 0;
// copy of current signature entry for matching processor
signatureType currentSignature;

unsigned long pagesize;
unsigned long pagemask;
unsigned long oldPage;
unsigned int progressBarCount;
// count errors
unsigned int errors;
bool gotEndOfFile;
unsigned long extendedAddress;

unsigned long lowestAddress;
unsigned long highestAddress;
unsigned long bytesWritten;
unsigned int lineCount;


/*
  LED codes (flash counts) - each sequence repeats 5 times

  If you are using the Crossroads programming board, the statuses are also shown
  as a 2-digit code in the 7-segment display. If you have the board with only one
  7-segment display you will see the first letter only.

  Problems with SD card or programming target chip
  ------------------------------------------------

  Red x 1 = Cannot open SD card (Sd)
  Red x 2 = Cannot read file 'firmware.hex' (FL)
  Red x 3 = Cannot enter programming mode in target chip (Ch)
  Red x 4 = This chip does not have a bootloader fuse (Fu)
  Red x 5 = Cannot read chip signature (Sg)
  Red x 6 = Unrecognized signature (Un)
  Red x 7 = Bad start address in hex file for code (Ad)
  Red x 8 = Verification error (bd)

  Problems with firmware.hex file
  -------------------------------

  Red + yellow x 1 = Line in file is too long (E1)
  Red + yellow x 2 = Line in file is too short (E2)
  Red + yellow x 3 = Line does not start with a colon (:) (E3)
  Red + yellow x 4 = Invalid hex digits (should be 0-9, A-F) (E4)
  Red + yellow x 5 = Bad sumcheck at end of line (E5)
  Red + yellow x 6 = Line not expected length (E6)
  Red + yellow x 7 = Unknown record type (E7)
  Red + yellow x 8 = No 'end of file' record in file (E8)
  Red + yellow x 9 = File will not fit into flash of target (LG)

  Worked OK
  ---------

  Green x 3 = Flashed and verified OK - sequence repeats 10 times (AA)

  Other statuses
  --------------

  Green (steady) = Ready to program (close switch to commence)
  Yellow (flashing) = Programming (Pr)
  Green (steady) + yellow (flashing) = verifying (uF)
  No LEDs on = No power?

*/
